// Experiment with Lighting

#include "Angel.h"
#include "Torus.h"


#ifdef __APPLE__
    #ifndef GLUT_3_2_CORE_PROFILE
        #define GLUT_3_2_CORE_PROFILE 0
        #define glBindVertexArray		glBindVertexArrayAPPLE
        #define glDeleteVertexArrays	glDeleteVertexArraysAPPLE
        #define glIsVertexArray			glIsVertexArrayAPPLE
        #define glGenVertexArrays  	glGenVertexArraysAPPLE
    #endif
#endif

//Define some geometry and normals here
typedef Angel::vec4 point4;

point4 cubeVerts[] = 
{
	point4( 0.5, 0.5, 0.5, 1), //0
	point4( 0.5, 0.5,-0.5, 1), //1
	point4( 0.5,-0.5, 0.5, 1), //2
	point4( 0.5,-0.5,-0.5, 1), //3
	point4(-0.5, 0.5, 0.5, 1), //4
	point4(-0.5, 0.5,-0.5, 1), //5
	point4(-0.5,-0.5, 0.5, 1), //6
	point4(-0.5,-0.5,-0.5, 1)  //7
};

point4 points[] = 
{
#define wireCubeStart 0
#define wireCubeVertices 30
	//Wire Cube - 30 vertices := 6 faces * (4 vertices per face + 1 to close loop)
	cubeVerts[0], cubeVerts[4], cubeVerts[6], cubeVerts[2], cubeVerts[0], //front
	cubeVerts[1], cubeVerts[0], cubeVerts[2], cubeVerts[3], cubeVerts[1], //right
	cubeVerts[5], cubeVerts[1], cubeVerts[3], cubeVerts[7], cubeVerts[5], //back 
	cubeVerts[4], cubeVerts[5], cubeVerts[7], cubeVerts[6], cubeVerts[4], //left 
	cubeVerts[4], cubeVerts[0], cubeVerts[1], cubeVerts[5], cubeVerts[4], //top
	cubeVerts[6], cubeVerts[7], cubeVerts[3], cubeVerts[2], cubeVerts[6], //bottom

#define solidCubeStart 30
#define solidCubeVertices 36
    //Solid Cube - 36 vertices := 6 faces * 2 Triangles per face * 3 vertices per triangle
	cubeVerts[0], cubeVerts[4], cubeVerts[6],  //front
	cubeVerts[0], cubeVerts[6], cubeVerts[2],  
	cubeVerts[1], cubeVerts[0], cubeVerts[2],  //right 
	cubeVerts[1], cubeVerts[2], cubeVerts[3],
	cubeVerts[5], cubeVerts[1], cubeVerts[3],  //back  
	cubeVerts[5], cubeVerts[3], cubeVerts[7],
	cubeVerts[4], cubeVerts[5], cubeVerts[7],  //left  
	cubeVerts[4], cubeVerts[7], cubeVerts[6],
	cubeVerts[4], cubeVerts[0], cubeVerts[1],  //top 
	cubeVerts[4], cubeVerts[1], cubeVerts[5],
	cubeVerts[6], cubeVerts[7], cubeVerts[3],  //bottom 
	cubeVerts[6], cubeVerts[3], cubeVerts[2],
	
}; 


vec4 right =	vec4( 1.0f, 0.0f, 0.0f, 0.0f);
vec4 left =		vec4(-1.0f, 0.0f, 0.0f, 0.0f);
vec4 top =		vec4( 0.0f, 1.0f, 0.0f, 0.0f);
vec4 bottom =	vec4( 0.0f,-1.0f, 0.0f, 0.0f);
vec4 front =	vec4( 0.0f, 0.0f, 1.0f, 0.0f);
vec4 back =		vec4( 0.0f, 0.0f,-1.0f, 0.0f);

vec4 white =	vec4( 1.0f, 1.0f, 1.0f, 1.0f);

vec4 normals[] = 
{
	//Normals for Wire Cube
	front, front, front, front, front,
	right, right, right, right, right,
	back, back, back, back, back, 
	left, left, left, left, left,
	top, top, top, top, top,
	bottom, bottom, bottom, bottom, bottom,
	
	//Normals for Solid Cube
	front, front, front, front, front, front,
	right, right, right, right, right, right,
	back, back, back, back, back, back,
	left, left, left, left, left, left,
	top, top, top, top, top, top,
	bottom, bottom, bottom, bottom, bottom, bottom
	
};


//Vertex array objects
GLuint cubes, torus;

//Shader Programs
GLuint program1, program2;

//Transformation Uniforms for Shader
GLint mvIndex;
GLint pIndex;

//Transformation Matrices
mat4 p, mv;

//Lighting Uniforms for Shader
GLint iLightPosition;
GLint ambientProduct;
GLint diffuseProduct;
GLint specularProduct;
GLint emission;
GLint shininess;
GLint lighting;
GLint uColor;

//Initial or "Default" Lighting Values
vec4 LightPosition(0.0f,0.0f,10.0f,0.0f);
vec4 LightAmbient(0.2f,0.2f,0.2f,1.0f);
vec4 LightDiffuse(1.0f,1.0f,1.0f,1.0f);
vec4 LightSpecular(1.0f,1.0f,1.0f,1.0f);

//Initial or "Default" Material Properties
vec4 MatAmbient(0.2f,0.2f,0.2f,1.0f);
vec4 MatDiffuse(0.8f,0.8f,0.8f,1.0f);
vec4 MatSpecular(0.0f,0.0f,0.0f,1.0f);
vec4 MatEmission(0.0f,0.0f,0.0f,1.0f);
GLfloat MatShininess = 5.0f;

//Number of vertices in Torus
GLuint ntVerts;

//Light rotation variable
GLint spin = 0;


//----------------------------------------------------------------------------

/* Initialize VAOs, shaders, material properties, lighting properties, and 
 * depth buffer.
 */
void init()
{	
	//Set up shaders and VAOs
	program1 = InitShader( "vShaderL4.glsl", "fShaderL4.glsl" );

	glUseProgram( program1 );
	glGenVertexArrays( 1, &cubes );
	glBindVertexArray( cubes );


	GLuint buffer;
	glGenBuffers( 1, &buffer );
	glBindBuffer( GL_ARRAY_BUFFER, buffer );
	glBufferData( GL_ARRAY_BUFFER, sizeof(points)+sizeof(normals),
		NULL, GL_STATIC_DRAW );
	glBufferSubData( GL_ARRAY_BUFFER, 0, sizeof(points), points );
	glBufferSubData( GL_ARRAY_BUFFER, sizeof(points), sizeof(normals), normals );
	GLuint vPosition = glGetAttribLocation( program1, "vPosition" );
	glEnableVertexAttribArray( vPosition );
	glVertexAttribPointer( vPosition, 4, GL_FLOAT, GL_FALSE, 0,
		BUFFER_OFFSET(0) );
	GLuint vNormal = glGetAttribLocation( program1, "vNormal" ); 
	glEnableVertexAttribArray( vNormal );

	glVertexAttribPointer( vNormal, 4, GL_FLOAT, GL_FALSE, 0,
		BUFFER_OFFSET(sizeof(points)) );


	//Get locations of transformation matrices from shader
 	mvIndex = glGetUniformLocation(program1, "mv");
	pIndex = glGetUniformLocation(program1, "p");

	//Get locations of lighting uniforms from shader
	iLightPosition = glGetUniformLocation(program1, "lightPosition");
	ambientProduct = glGetUniformLocation(program1, "ambientProduct");
	diffuseProduct = glGetUniformLocation(program1, "diffuseProduct");
	specularProduct = glGetUniformLocation(program1, "specularProduct");
	emission = glGetUniformLocation(program1, "emission");
	shininess = glGetUniformLocation(program1, "shininess");
	lighting = glGetUniformLocation(program1, "lighting");
	uColor = glGetUniformLocation(program1, "uColor");

	//Set default lighting and material properties in shader.
	glUniform4fv(iLightPosition, 1, LightPosition);
	glUniform4fv(ambientProduct, 1, MatAmbient*LightAmbient);
	glUniform4fv(diffuseProduct, 1, MatDiffuse*LightDiffuse);
	glUniform4fv(specularProduct, 1, MatSpecular*LightSpecular);
	glUniform4fv(emission, 1, MatEmission);
	glUniform1f(shininess, MatShininess);
	glUniform1f(lighting, GL_TRUE);
	glUniform4f(uColor, 1.0f, 1.0f, 1.0f, 1.0f);

	//Create a Torus VAO. Read function description in Torus.h for details.
	createTorusVAO(0.275f, 0.85f, 10, 20, vPosition, vNormal, torus, ntVerts);
	glEnable(GL_DEPTH_TEST);
}

//----------------------------------------------------------------------------

/* Here the light position is modified by a modeling transformation stored in
 * the mat4 called light which is based on a rotated version of the mv matrix.
 * This places the light at a position in world coordinates.  A cube
 * represents the position of the light. 
 */
void display( void )
{
	//Clear the drawing buffer
	glClear( GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT );

	mv = mat4();

	//set up viewing transformation
	point4 eye(0,0,5,1);
	point4 at(0,0,0,1);
	vec4   up(0,1,0,0);

	//mv will represent the center of the scene
	mv *= LookAt(eye, at, up);

	//Position light relative to scene
	mat4 light = mv;
	light *= RotateX( spin );
    vec4 position( 0.0, 0.0, 1.5, 1.0 );
	glUniform4fv(iLightPosition, 1,  light*position);

	//Draw small unlit wire cube at light's position
	light *= Translate(0,0,1.5);					//Move to light's position
	light *= Scale(0.1f, 0.1f, 0.1f);				//Make box small
	glUniformMatrix4fv(mvIndex, 1, GL_TRUE, light); //Apply transformations
	glBindVertexArray(cubes);						//Bind VAO with box definitions
	glUniform1f(lighting, GL_FALSE);				//Turn off lighting
	glUniform4f(uColor, 1.0f, 1.0f, 1.0f, 1.0f);	//Set the unlit color
	glDrawArrays(GL_LINE_STRIP, wireCubeStart, wireCubeVertices);
	glUniform1f(lighting, GL_TRUE);					//Turn on lighting

	//Draw Torus at center of scene
	glUniformMatrix4fv(mvIndex, 1, GL_TRUE, mv);	//Reset to center of scene
	glBindVertexArray(torus);						//Bind VAO with torus
	glDrawArrays(GL_TRIANGLES, 0, ntVerts);			//Draw Torus

	//Show what was drawn
	glutSwapBuffers();
}

//----------------------------------------------------------------------------

void reshape (int w, int h)
{
	GLfloat aspect = GLfloat(w)/GLfloat(h);
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	p = mat4();
	p *= Perspective(40.0f, aspect,  1.0f, 20.0f);
	glUniformMatrix4fv(pIndex, 1, GL_TRUE, p);
}


void keyboard( unsigned char key, int x, int y )
{
	switch (key)
	{
	case '1':
		//Switch to vertex based lighting
		break;
	case '2':
		//Switch to fragment based lighting
		break;

	case 033:  // Escape key
	case 'q': case 'Q':
		exit( EXIT_SUCCESS );
		break;
	}
}

//----------------------------------------------------------------------------

void mouse(int button, int state, int x, int y)
{
   switch (button) {
      case GLUT_LEFT_BUTTON:
         if (state == GLUT_DOWN) {
            spin = (spin + 30) % 360;
            glutPostRedisplay();
         }
         break;
      default:
         break;
   }
}

//----------------------------------------------------------------------------

int main( int argc, char **argv )
{
	glutInit( &argc, argv );
	glutInitWindowSize( 512, 512 );
    
#ifdef __APPLE__
	glutInitDisplayMode( GLUT_3_2_CORE_PROFILE | GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
#else
	// If you are using freeglut, the next two lines will check if
	// the code is truly 3.2. Otherwise, comment them out
	glutInitContextVersion( 3, 2 );
	glutInitContextProfile( GLUT_CORE_PROFILE );
	glutInitDisplayMode( GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH );
    
#endif
    
    glutCreateWindow( "Lab 4 Lighting Exercise" );
    glewExperimental = GL_TRUE;
	glewInit();

	init();
	glutDisplayFunc( display );
	glutKeyboardFunc( keyboard );
	glutReshapeFunc( reshape );
    glutMouseFunc(mouse);

	glutMainLoop();
	return 0;
}
