#version 120

attribute  vec4 vPosition;
attribute  vec4 vNormal;

uniform mat4 p; 
uniform mat4 mv;
uniform float shininess;
uniform bool lighting;
uniform vec4 ambientProduct;
uniform vec4 diffuseProduct;
uniform vec4 specularProduct;
uniform vec4 emission;
uniform vec4 lightPosition;
uniform vec4 uColor;

varying vec4 color;

vec4 mvPosition;

void main() 
{

  //Transform the point
  mvPosition = mv*vPosition;  //Point's position in view space is used often, so keep intermidiate result
  gl_Position = p*mvPosition; 

  //Set up Normal, Light, Eye and Half vectors
  vec3 N = normalize((mv*vNormal).xyz);
  vec3 L = normalize(lightPosition.xyz - mvPosition.xyz);
  if (lightPosition.w == 0.0) L = normalize(lightPosition.xyz);
  vec3 E = -normalize(mvPosition.xyz);
  vec3 H = normalize(L+E); 

  //Calculate diffuse coefficient
  float Kd = max(dot(L,N), 0.0);

  //Calculate Blinn-Phong specular coefficient
  float Ks = pow(max(dot(N,H), 0.0), shininess);

  //Calculate lit colour for this pixel
  color = Kd * diffuseProduct + Ks * specularProduct + ambientProduct + emission;
  if (lighting == false) 
	color = uColor;
}
